# coding=utf-8
"""
DNSCOM API
DNS.COM 接口解析操作库
http://open.dns.com/
@author: Bigjin
@mailto: i@bigjin.com
"""

from hashlib import md5
from json import loads as jsondecode
from logging import debug, info, warning
from time import mktime
from datetime import datetime

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode


__author__ = 'Bigjin'
# __all__ = ["request", "ID", "TOKEN", "PROXY"]


class Config:
    ID = "id"
    TOKEN = "TOKEN"
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SITE = "www.dns.com"  # API endpoint
    METHOD = "POST"  # 请求方法


def signature(params):
    """
    计算签名,返回签名后的查询参数
    """
    params.update({
        'apiKey': Config.ID,
        'timestamp': mktime(datetime.now().timetuple()),
    })
    query = urlencode(sorted(params.items()))
    debug(query)
    sign = query
    debug("signString: %s", sign)

    sign = md5((sign + Config.TOKEN).encode('utf-8')).hexdigest()
    params["hash"] = sign

    return params


def request(action, param=None, **params):
    """
    发送请求数据
    """
    if param:
        params.update(param)
    params = dict((k, params[k]) for k in params if params[k] is not None)
    params = signature(params)
    info("%s/api/%s/ : params:%s", API.SITE, action, params)

    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)

    conn.request(API.METHOD, '/api/' + action + '/', urlencode(params),
                 {"Content-type": "application/x-www-form-urlencoded"})
    response = conn.getresponse()
    result = response.read().decode('utf8')
    conn.close()

    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]:%s', action, response.status, result)
        raise Exception(result)
    else:
        data = jsondecode(result)
        debug('%s : result:%s', action, data)
        if data.get('code') != 0:
            raise Exception("api error:", data.get('message'))
        data = data.get('data')
        if data is None:
            raise Exception('response data is none')
        return data


def get_domain_info(domain):
    """
    切割域名获取主域名和对应ID
    """
    if len(domain.split('.')) > 2:
        domains = domain.split('.', 1)
        sub = domains[0]
        main = domains[1]
    else:
        sub = ''  # 接口有bug 不能传 @ * 作为主机头，但是如果为空，默认为 @
        main = domain

    res = request("domain/getsingle", domainID=main)
    domain_id = res.get('domainID')
    return sub, main, domain_id


def get_records(domain, domain_id, **conditions):
    """
        获取记录ID
        返回满足条件的所有记录[]
        TODO 大于500翻页
    """
    if not hasattr(get_records, "records"):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ("recordID", "record", "type", "viewID",
                            "TTL", "state", "value")

    if not domain in get_records.records:
        get_records.records[domain] = {}
        data = request("record/list",
                       domainID=domain_id, pageSize=500)
        if data.get('data'):
            for record in data.get('data'):
                get_records.records[domain][record["recordID"]] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}
    records = {}
    for (rid, record) in get_records.records[domain].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[rid] = record
    return records


def update_record(domain, value, record_type='A'):
    """
        更新记录
    """
    info(">>>>>%s(%s)", domain, record_type)
    sub, main, domain_id = get_domain_info(domain)

    records = get_records(main, domain_id, record=sub, type=record_type)
    result = {}

    if records:
        for (rid, record) in records.items():
            if record["value"] != value:
                debug(sub, record)
                res = request("record/modify", domainID=domain_id,
                              recordID=rid, newvalue=value, newTTL=Config.TTL)
                if res:
                    # update records
                    get_records.records[main][rid]["value"] = value
                    result[rid] = res
                else:
                    result[rid] = "update fail!\n" + str(res)
            else:
                result[rid] = domain
    else:
        res = request("record/create", domainID=domain_id,
                      value=value, host=sub, type=record_type, TTL=Config.TTL)
        if res:
            # update records INFO
            rid = res.get('recordID')
            get_records.records[main][rid] = {
                'value': value,
                "recordID": rid,
                "record": sub,
                "type": record_type
            }
            result = res
        else:
            result = domain + " created fail!"
    return result



# coding=utf-8
"""
HuaweiDNS API
华为DNS解析操作库
https://support.huaweicloud.com/api-dns/zh-cn_topic_0037134406.html
@author: cybmp3
"""

import hashlib
import hmac
import binascii
import json
from uuid import uuid4
from base64 import b64encode

#from json import loads as jsondecode

from logging import debug, info, warning
from datetime import datetime

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode, quote_plus, quote
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode, quote_plus, quote

__author__ = 'New Future'
BasicDateFormat = "%Y%m%dT%H%M%SZ"
Algorithm = "SDK-HMAC-SHA256"


# __all__ = ["request", "ID", "TOKEN", "PROXY"]


class Config:
    ID = "id"  # AK
    TOKEN = "TOKEN"  # AS
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SCHEME = 'https'
    SITE = 'dns.myhuaweicloud.com'  # API endpoint


def HexEncodeSHA256Hash(data):
    sha256 = hashlib.sha256()
    sha256.update(data)
    return sha256.hexdigest()


def StringToSign(canonical_request, t):
    b = HexEncodeSHA256Hash(canonical_request)
    return "%s\n%s\n%s" % (Algorithm, datetime.strftime(t, BasicDateFormat), b)


def CanonicalHeaders(headers, signed_headers):
    a = []
    __headers = {}
    for key in headers:
        key_encoded = key.lower()
        value = headers[key]
        value_encoded = value.strip()
        __headers[key_encoded] = value_encoded
    for key in signed_headers:
        a.append(key + ":" + __headers[key])
    return '\n'.join(a) + "\n"


def request(method, path, param=None, body=None, **params):
    # path 是不带host但是 前面需要带 / , body json 字符串或者自己从dict转换下
    # 也可以自己改成 判断下是不是post 是post params就是body
    if param:
        params.update(param)

    query = urlencode(sorted(params.items()))
    headers = {"content-type": "application/json"}  # 初始化header
    headers["X-Sdk-Date"] = datetime.strftime(datetime.utcnow(), BasicDateFormat)
    headers["host"] = API.SITE
    # 如何后来有需要把header头 key转换为小写 value 删除前导空格和尾随空格
    sign_headers = []
    for key in headers:
        sign_headers.append(key.lower())
    # 先排序
    sign_headers.sort()

    if body is None:
        body = ""

    hex_encode = HexEncodeSHA256Hash(body.encode('utf-8'))
    # 生成文档中的CanonicalRequest
    canonical_headers = CanonicalHeaders(headers, sign_headers)

    # 签名中的path 必须 / 结尾
    if path[-1] != '/':
        sign_path = path + "/"
    else:
        sign_path = path

    canonical_request = "%s\n%s\n%s\n%s\n%s\n%s" % (method.upper(), sign_path, query,
                                                    canonical_headers, ";".join(sign_headers), hex_encode)

    hashed_canonical_request = HexEncodeSHA256Hash(canonical_request.encode('utf-8'))

    # StringToSign
    str_to_sign = "%s\n%s\n%s" % (Algorithm, headers['X-Sdk-Date'], hashed_canonical_request)

    secret = Config.TOKEN
    # 计算签名  HexEncode(HMAC(Access Secret Key, string to sign))
    signature = hmac.new(secret.encode('utf-8'), str_to_sign.encode('utf-8'), digestmod=hashlib.sha256).digest()
    signature = binascii.hexlify(signature).decode()
    # 添加签名信息到请求头
    auth_header = "%s Access=%s, SignedHeaders=%s, Signature=%s" % (
        Algorithm, Config.ID, ";".join(sign_headers), signature)
    headers['Authorization'] = auth_header
    # 创建Http请求

    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)
    conn.request(method, API.SCHEME + "://" + API.SITE + path + '?' + query, body, headers)
    info(API.SCHEME + "://" + API.SITE + path + '?' + query, body)
    resp = conn.getresponse()
    data = resp.read().decode('utf8')
    resp.close()
    if resp.status < 200 or resp.status >= 300:

        warning('%s : error[%d]: %s', path, resp.status, data)
        raise Exception(data)
    else:
        data = json.loads(data)
        debug('%s : result:%s', path, data)
        return data


def get_zone_id(domain):
    """
    切割域名获取主域名和对应ID https://support.huaweicloud.com/api-dns/zh-cn_topic_0037134402.html
    """
    zones = request('GET', '/v2/zones', limit=500)['zones']
    domain += '.'
    zone = next((z for z in zones if domain.endswith(z.get('name'))), None)
    zoneid = zone and zone['id']
    return zoneid

def get_records(zoneid, **conditions):
    """
        获取记录ID
        返回满足条件的所有记录[]
        https://support.huaweicloud.com/api-dns/zh-cn_topic_0037129970.html
        TODO 大于500翻页
    """
    cache_key = zoneid + "_" + \
                conditions.get('name', "") + "_" + conditions.get('type', "")
    if not hasattr(get_records, 'records'):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ('id', 'type', 'name', 'records', 'ttl')

    if not zoneid in get_records.records:
        get_records.records[cache_key] = {}

        data = request('GET', '/v2/zones/' + zoneid + '/recordsets',
                       limit=500, **conditions)

        # https://{DNS_Endpoint}/v2/zones/2c9eb155587194ec01587224c9f90149/recordsets?limit=&offset=
        if data:
            for record in data['recordsets']:
                info(record)
                get_records.records[cache_key][record['id']] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}
    records = {}
    for (zid, record) in get_records.records[cache_key].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[zid] = record
    return records


def update_record(domain, value, record_type='A', name=None):
    """
        更新记录
        update
        https://support.huaweicloud.com/api-dns/dns_api_64006.html
        add
        https://support.huaweicloud.com/api-dns/zh-cn_topic_0037134404.html
    """
    info(">>>>>%s(%s)", domain, record_type)
    zoneid = get_zone_id(domain)
    if not zoneid:
        raise Exception("invalid domain: [ %s ] " % domain)
    domain += '.'
    records = get_records(zoneid, name=domain, type=record_type)
    cache_key = zoneid + "_" + domain + "_" + record_type
    result = {}
    if records:  # update
        for (rid, record) in records.items():
            if record['records'] != value:
                """
                {
                    "description": "This is an example record set.",
                    "ttl": 3600,
                    "records": [
                        "192.168.10.1",
                        "192.168.10.2"
                    ]
                }
                """
                body = {
                    "records":[
                        value
                    ]
                }
                res = request('PUT', '/v2/zones/' + zoneid + '/recordsets/' + record['id'],
                              body=str(json.dumps(body)))
                if res:
                    get_records.records[cache_key][rid]['records'] = value
                    result[rid] = res.get("name")
                else:
                    result[rid] = "Update fail!\n" + str(res)
            else:
                result[rid] = domain
    else:  # create
        print(domain)
        body = {
            "name": domain,
            "type": record_type,
            "records":[
                value
            ]
        }
        res = request('POST', '/v2/zones/' + zoneid + '/recordsets',
                      body=str(json.dumps(body)))
        if res:
            get_records.records[cache_key][res['id']] = res
            result = res
        else:
            result = domain + " created fail!"
    return result



# coding=utf-8
"""
DNSPOD API
DNSPOD 接口解析操作库
http://www.dnspod.cn/docs/domains.html
@author: New Future
"""

from json import loads as jsondecode
from logging import debug, info, warning
from os import environ
try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode

__author__ = 'New Future'


class Config:
    ID = "token id"
    TOKEN = "token key"
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SITE = "dnsapi.cn"  # API endpoint
    METHOD = "POST"  # 请求方法
    TOKEN_PARAM = "login_token"  # token参数
    DEFAULT = "默认"  # 默认线路名


def request(action, param=None, **params):
    """
    发送请求数据
    """
    if param:
        params.update(param)
    params = dict((k, params[k]) for k in params if params[k] is not None)
    params.update({API.TOKEN_PARAM: '***', 'format': 'json'})
    info("%s/%s : %s", API.SITE, action, params)
    params[API.TOKEN_PARAM] = "%s,%s" % (Config.ID, Config.TOKEN)
    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)

    conn.request(API.METHOD, '/' + action, urlencode(params), {
        "Content-type": "application/x-www-form-urlencoded",
        "User-Agent": "DDNS/%s (ddns@newfuture.cc)" % environ.get("DDNS_VERSION", "1.0.0")
    })
    response = conn.getresponse()
    res = response.read().decode('utf8')
    conn.close()

    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]:%s', action, response.status, res)
        raise Exception(res)
    else:
        data = jsondecode(res)
        debug('%s : result:%s', action, data)
        if not data:
            raise Exception("empty response")
        elif data.get("status", {}).get("code") == "1":
            return data
        else:
            raise Exception(data.get('status', {}))


def get_domain_info(domain):
    """
    切割域名获取主域名和对应ID
    """
    domain_split = domain.split('.')
    sub, did = None, None
    main = domain_split.pop()
    while domain_split:  # 通过API判断,最后两个，三个递增
        main = domain_split.pop() + '.' + main
        did = get_domain_id(main)
        if did:
            sub = ".".join(domain_split) or '@'
            # root domain根域名https://github.com/NewFuture/DDNS/issues/9
            break
    info('domain_id: %s, sub: %s', did, sub)
    return did, sub


def get_domain_id(domain):
    """
        获取域名ID
        http://www.dnspod.cn/docs/domains.html#domain-info
    """
    if not hasattr(get_domain_id, "domain_list"):
        get_domain_id.domain_list = {}  # "静态变量"存储已查询过的id

    if domain in get_domain_id.domain_list:
        # 如果已经存在直接返回防止再次请求
        return get_domain_id.domain_list[domain]
    else:
        try:
            info = request('Domain.Info', domain=domain)
        except Exception:
            return
        did = info.get("domain", {}).get("id")
        if did:
            get_domain_id.domain_list[domain] = did
            return did


def get_records(did, **conditions):
    """
        获取记录ID
        返回满足条件的所有记录[]
        TODO 大于3000翻页
        http://www.dnspod.cn/docs/records.html#record-list
    """
    if not hasattr(get_records, "records"):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ("id", "name", "type", "line",
                            "line_id", "enabled", "mx", "value")

    if not did in get_records.records:
        get_records.records[did] = {}
        data = request('Record.List', domain_id=did)
        if data:
            for record in data.get('records'):
                get_records.records[did][record["id"]] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}

    records = {}
    for (did, record) in get_records.records[did].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[did] = record
    return records


def update_record(domain, value, record_type="A"):
    """
    更新记录
    """
    info(">>>>>%s(%s)", domain, record_type)
    domainid, sub = get_domain_info(domain)
    if not domainid:
        raise Exception("invalid domain: [ %s ] " % domain)

    records = get_records(domainid, name=sub, type=record_type)
    result = {}
    if records:  # update
        # http://www.dnspod.cn/docs/records.html#record-modify
        for (did, record) in records.items():
            if record["value"] != value:
                debug(sub, record)
                res = request('Record.Modify', record_id=did, record_line=record["line"].replace("Default", "default").encode(
                    "utf-8"), value=value, sub_domain=sub, domain_id=domainid, record_type=record_type, ttl=Config.TTL)
                if res:
                    get_records.records[domainid][did]["value"] = value
                    result[did] = res.get("record")
                else:
                    result[did] = "update fail!\n" + str(res)
            else:
                result[did] = domain
    else:  # create
        # http://www.dnspod.cn/docs/records.html#record-create
        res = request("Record.Create", domain_id=domainid, value=value,
                      sub_domain=sub, record_type=record_type, record_line=API.DEFAULT, ttl=Config.TTL)
        if res:
            did = res.get("record")["id"]
            get_records.records[domainid][did] = res.get("record")
            get_records.records[domainid][did].update(
                value=value, sub_domain=sub, record_type=record_type)
            result = res.get("record")
        else:
            result = domain + " created fail!"
    return result



# coding=utf-8
"""
DNSPOD API
DNSPOD 接口解析操作库
http://www.dnspod.com/docs/domains.html
@author: New Future
"""

from dns.dnspod import *

API.SITE = "api.dnspod.com"
API.DEFAULT = "default"
API.TOKEN_PARAM = "user_token"



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
DDNS
@author: New Future
@modified: rufengsuixing
"""
from __future__ import print_function
from argparse import ArgumentParser, RawTextHelpFormatter
from json import load as loadjson, dump as dumpjson
from time import ctime
from os import path, environ, stat, name as os_name
from tempfile import gettempdir
from logging import DEBUG, basicConfig, info, warning, error
from subprocess import check_output

import sys

from util import ip
from util.cache import Cache

__version__ = "v2.9.1@2020-03-26T13:04:46+00:00"  # CI 时会被Tag替换
__description__ = "automatically update DNS records to dynamic local IP [自动更新DNS记录指向本地IP]"
__doc__ = """
ddns[%s]
(i) homepage or docs [文档主页]: https://ddns.newfuture.cc/
(?) issues or bugs [问题和帮助]: https://github.com/NewFuture/DDNS/issues
Copyright (c) New Future (MIT License)
""" % (__version__)

environ["DDNS_VERSION"] = "v2.9.1"

if getattr(sys, 'frozen', False):
    # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-OpenSSL-Certificate
    environ['SSL_CERT_FILE'] = path.join(
        getattr(sys, '_MEIPASS'), 'lib', 'cert.pem')

CACHE_FILE = path.join(gettempdir(), 'ddns.cache')


def get_config(key=None, default=None, path="config.json"):
    """
    读取配置
    """
    if not hasattr(get_config, "config"):
        try:
            with open(path) as configfile:
                get_config.config = loadjson(configfile)
                get_config.time = stat(path).st_mtime
        except IOError:
            error(' Config file `%s` does not exist!' % path)
            with open(path, 'w') as configfile:
                configure = {
                    "$schema": "https://ddns.newfuture.cc/schema/v2.8.json",
                    "id": "YOUR ID or EAMIL for DNS Provider",
                    "token": "YOUR TOKEN or KEY for DNS Provider",
                    "dns": "dnspod",
                    "ipv4": [
                        "newfuture.cc",
                        "ddns.newfuture.cc"
                    ],
                    "ipv6": [
                        "newfuture.cc",
                        "ipv6.ddns.newfuture.cc"
                    ],
                    "index4": "default",
                    "index6": "default",
                    "ttl": None,
                    "proxy": None,
                    "debug": False,
                }
                dumpjson(configure, configfile, indent=2, sort_keys=True)
            sys.exit("New template configure file `%s` is generated." % path)
        except:
            sys.exit('fail to load config from file: %s' % path)
    if key:
        return get_config.config.get(key, default)
    else:
        return get_config.config


def get_ip(ip_type):
    """
    get IP address
    """
    index = get_config('index' + ip_type, "default")
    if index is False:  # disabled
        return False
    elif str(index).isdigit():  # 数字 local eth
        value = getattr(ip, "local_v" + ip_type)(index)
    elif index.startswith('cmd:'):  # cmd
        value = str(check_output(index[4:]).strip().decode('utf-8'))
    elif index.startswith('shell:'):  # shell
        value = str(check_output(
            index[6:], shell=True).strip().decode('utf-8'))
    elif index.startswith('url:'):  # 自定义 url
        value = getattr(ip, "public_v" + ip_type)(index[4:])
    elif index.startswith('regex:'):  # 正则 regex
        value = getattr(ip, "regex_v" + ip_type)(index[6:])
    elif any((c in index) for c in '*.:'):  # 兼容 regex
        value = getattr(ip, "regex_v" + ip_type)(index)
    else:
        value = getattr(ip, index + "_v" + ip_type)()

    return value


def change_dns_record(dns, proxy_list, **kw):
    for proxy in proxy_list:
        if not proxy or (proxy.upper() in ['DIRECT', 'NONE']):
            dns.PROXY = None
        else:
            dns.PROXY = proxy
        record_type, domain = kw['record_type'], kw['domain']
        print('\n%s(%s) ==> %s [via %s]' %
              (domain, record_type, kw['ip'], proxy))
        try:
            return dns.update_record(domain, kw['ip'], record_type=record_type)
        except Exception as e:
            error(e)
    return False


def update_ip(ip_type, cache, dns, proxy_list):
    """
    更新IP
    """
    ipname = 'ipv' + ip_type
    domains = get_config(ipname)
    if not domains:
        return None
    address = get_ip(ip_type)
    if not address:
        error('Fail to get %s address!' ,ipname)
        return False
    elif cache and (address == cache[ipname]):
        print('.', end=" ")  # 缓存命中
        return True
    record_type = (ip_type == '4') and 'A' or 'AAAA'
    update_fail = False  # https://github.com/NewFuture/DDNS/issues/16
    for domain in domains:
        if change_dns_record(dns, proxy_list, domain=domain, ip=address, record_type=record_type):
            update_fail = True
    if cache is not False:
        # 如果更新失败删除缓存
        cache[ipname] = update_fail and address


def main():
    """
    更新
    """
    parser = ArgumentParser(description=__description__,
                            epilog=__doc__, formatter_class=RawTextHelpFormatter)
    parser.add_argument('-v', '--version',
                        action='version', version=__version__)
    parser.add_argument('-c', '--config',
                        default="config.json", help="run with config file [配置文件路径]")
    config_file = parser.parse_args().config
    get_config(path=config_file)
    # Dynamicly import the dns module as configuration
    dns_provider = str(get_config('dns', 'dnspod').lower())
    dns = getattr(__import__('dns', fromlist=[dns_provider]), dns_provider)
    dns.Config.ID = get_config('id')
    dns.Config.TOKEN = get_config('token')
    dns.Config.TTL = get_config('ttl')
    if get_config('debug'):
        ip.DEBUG = get_config('debug')
        basicConfig(
            level=DEBUG,
            format='%(asctime)s <%(module)s.%(funcName)s> %(lineno)d@%(pathname)s \n[%(levelname)s] %(message)s')
        print("DDNS[", __version__, "] run:", os_name, sys.platform)
        print("Configuration was loaded from <==", path.abspath(config_file))
        print("=" * 25, ctime(), "=" * 25, sep=' ')

    proxy = get_config('proxy') or 'DIRECT'
    proxy_list = proxy.strip('; ') .split(';')

    cache = get_config('cache', True) and Cache(CACHE_FILE)
    if cache is False:
        info("Cache is disabled!")
    elif len(cache) < 1 or get_config.time >= cache.time:
        warning("Cache file is out of dated.")
        cache.clear()
    update_ip('4', cache, dns, proxy_list)
    update_ip('6', cache, dns, proxy_list)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
cache module
文件缓存
"""


from os import path, stat
from pickle import dump, load
from time import time

from logging import info, debug, error

try:
    from collections.abc import MutableMapping
except ImportError:
    # Python 2 imports
    from collections import MutableMapping


class Cache(MutableMapping):
    """
    using file to Cache data as dictionary
    """

    def __init__(self, path, sync=False):
        self.__data = {}
        self.__filename = path
        self.__sync = sync
        self.__time = time()
        self.__changed = False
        self.load()

    @property
    def time(self):
        """
        缓存修改时间
        """
        return self.__time

    def load(self, file=None):
        """
        load data from path
        """
        if not file:
            file = self.__filename

        debug('load cache data from %s', file)
        if path.isfile(file):
            with open(self.__filename, 'rb') as data:
                try:
                    self.__data = load(data)
                    self.__time = stat(file).st_mtime
                except ValueError:
                    self.__data = {}
                    self.__time = time()
                except Exception as e:
                    error(e)
                    self.__data = {}
                    self.__time = time()
        else:
            info('cache file not exist')
            self.__data = {}
            self.__time = time()
        return self

    def data(self, key=None, default=None):
        """
        获取当前字典或者制定得键值
        """
        if self.__sync:
            self.load()

        if key is None:
            return self.__data
        else:
            return self.__data.get(key, default)

    def sync(self):
        """Sync the write buffer with the cache files and clear the buffer.
        """
        if self.__changed:
            with open(self.__filename, 'wb') as data:
                dump(self.__data, data)
                debug('save cache data to %s', self.__filename)
            self.__time = time()
            self.__changed = False
        return self

    def close(self):
        """Sync the write buffer, then close the cache.
        If a closed :class:`FileCache` object's methods are called, a
        :exc:`ValueError` will be raised.
        """
        self.sync()
        del self.__data
        del self.__filename
        del self.__time
        self.__sync = False

    def __update(self):
        self.__changed = True
        if self.__sync:
            self.sync()
        else:
            self.__time = time()

    def clear(self):
        if self.data():
            self.__data = {}
            self.__update()

    def __setitem__(self, key, value):
        if self.data(key) != value:
            self.__data[key] = value
            self.__update()

    def __delitem__(self, key):
        if key in self.data():
            del self.__data[key]
            self.__update()

    def __getitem__(self, key):
        return self.data(key)

    def __iter__(self):
        for key in self.data():
            yield key

    def __len__(self):
        return len(self.data())

    def __contains__(self, key):
        return key in self.data()

    def __str__(self):
        return self.data().__str__()

    def __del__(self):
        self.close()



# coding=utf-8
"""
AliDNS API
阿里DNS解析操作库
https://help.aliyun.com/document_detail/29739.html
@author: New Future
"""

from hashlib import sha1
from hmac import new as hmac
from uuid import uuid4
from base64 import b64encode
from json import loads as jsondecode
from logging import debug, info, warning
from datetime import datetime

try:
    # python 2
    from httplib import HTTPSConnection
    from urllib import urlencode, quote_plus, quote
except ImportError:
    # python 3
    from http.client import HTTPSConnection
    from urllib.parse import urlencode, quote_plus, quote

__author__ = 'New Future'
# __all__ = ["request", "ID", "TOKEN", "PROXY"]


class Config:
    ID = "id"
    TOKEN = "TOKEN"
    PROXY = None  # 代理设置
    TTL = None


class API:
    # API 配置
    SITE = "alidns.aliyuncs.com"  # API endpoint
    METHOD = "POST"  # 请求方法


def signature(params):
    """
    计算签名,返回签名后的查询参数
    """
    params.update({
        'Format': 'json',
        'Version': '2015-01-09',
        'AccessKeyId': Config.ID,
        'Timestamp': datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ'),
        'SignatureMethod': 'HMAC-SHA1',
        'SignatureNonce': uuid4(),
        'SignatureVersion': "1.0",
    })
    query = urlencode(sorted(params.items()))
    debug(query)
    sign = API.METHOD + "&" + quote_plus("/") + "&" + quote(query, safe='')
    debug("signString: %s", sign)

    sign = hmac((Config.TOKEN + "&").encode('utf-8'),
                sign.encode('utf-8'), sha1).digest()
    sign = b64encode(sign).strip()
    params["Signature"] = sign
    return params


def request(param=None, **params):
    """
    发送请求数据
    """
    if param:
        params.update(param)
    params = dict((k, params[k]) for k in params if params[k] is not None)
    params = signature(params)
    info("%s: %s", API.SITE, params)

    if Config.PROXY:
        conn = HTTPSConnection(Config.PROXY)
        conn.set_tunnel(API.SITE, 443)
    else:
        conn = HTTPSConnection(API.SITE)
    conn.request(API.METHOD, '/', urlencode(params),
                 {"Content-type": "application/x-www-form-urlencoded"})
    response = conn.getresponse()
    data = response.read().decode('utf8')
    conn.close()

    if response.status < 200 or response.status >= 300:
        warning('%s : error[%d]: %s', params['Action'], response.status, data)
        raise Exception(data)
    else:
        data = jsondecode(data)
        debug('%s : result:%s', params['Action'], data)
        return data


def get_domain_info(domain):
    """
    切割域名获取主域名和对应ID
    https://help.aliyun.com/document_detail/29755.html
    http://alidns.aliyuncs.com/?Action=GetMainDomainName&InputString=www.example.com
    """
    res = request(Action="GetMainDomainName", InputString=domain)
    sub, main = res.get('RR'), res.get('DomainName')
    return sub, main


def get_records(domain, **conditions):
    """
        获取记录ID
        返回满足条件的所有记录[]
        https://help.aliyun.com/document_detail/29776.html
        TODO 大于500翻页
    """
    if not hasattr(get_records, "records"):
        get_records.records = {}  # "静态变量"存储已查询过的id
        get_records.keys = ("RecordId", "RR", "Type", "Line",
                            "Locked", "Status", "Priority", "Value")

    if not domain in get_records.records:
        get_records.records[domain] = {}
        data = request(Action="DescribeDomainRecords",
                       DomainName=domain, PageSize=500)
        if data:
            for record in data.get('DomainRecords').get('Record'):
                get_records.records[domain][record["RecordId"]] = {
                    k: v for (k, v) in record.items() if k in get_records.keys}
    records = {}
    for (rid, record) in get_records.records[domain].items():
        for (k, value) in conditions.items():
            if record.get(k) != value:
                break
        else:  # for else push
            records[rid] = record
    return records


def update_record(domain, value, record_type='A'):
    """
        更新记录
        update
        https://help.aliyun.com/document_detail/29774.html
        add
        https://help.aliyun.com/document_detail/29772.html?
    """
    debug(">>>>>%s(%s)", domain, record_type)
    sub, main = get_domain_info(domain)
    if not sub:
        raise Exception("invalid domain: [ %s ] " % domain)

    records = get_records(main, RR=sub, Type=record_type)
    result = {}

    if records:
        for (rid, record) in records.items():
            if record["Value"] != value:
                debug(sub, record)
                res = request(Action="UpdateDomainRecord", RecordId=rid,
                              Value=value, RR=sub, Type=record_type, TTL=Config.TTL)
                if res:
                    # update records
                    get_records.records[main][rid]["Value"] = value
                    result[rid] = res
                else:
                    result[rid] = "update fail!\n" + str(res)
            else:
                result[rid] = domain
    else:  # https://help.aliyun.com/document_detail/29772.html
        res = request(Action="AddDomainRecord", DomainName=main,
                      Value=value, RR=sub, Type=record_type, TTL=Config.TTL)
        if res:
            # update records INFO
            rid = res.get('RecordId')
            get_records.records[main][rid] = {
                'Value': value,
                "RecordId": rid,
                "RR": sub,
                "Type": record_type
            }
            result = res
        else:
            result = domain + " created fail!"
    return result
